use crate::stack_vec::StackVec;

#[derive(Debug, PartialEq, Eq)]
pub struct Ax25Frame<const N: usize> {
    source: StackVec<7, u8>,
    dest: StackVec<7, u8>,
    info: StackVec<N, u8>,
}

impl<const N: usize> Ax25Frame<N> {
    pub fn new(source: &[u8], dest: &[u8], info: &[u8]) -> Option<Self> {
        let source = StackVec::try_from(source).ok()?;
        let dest = StackVec::try_from(dest).ok()?;
        let info = StackVec::try_from(info).ok()?;

        Some(Self { source, dest, info })
    }

    pub fn decode_raw(raw: &[u8]) -> Option<Self> {
        let mut iter = raw.iter();

        // address
        let mut address: StackVec<14, u8> = StackVec::new();
        loop {
            let mut byte = *iter.next()?;
            let finished = byte & 1;
            byte >>= 1;
            let has_space = address.push(byte).is_some();

            if finished == 1 || !has_space {
                break;
            }
        }

        // technically we can also have addresses of length 28
        // but I haven't bothered to implement them.
        if address.len() != 14 {
            return None;
        }

        let mut dest = StackVec::new();
        for &b in &address.as_slice()[0..6] {
            // omit spaces
            if b != 32 {
                dest.push(b)?;
            }
        }

        let mut source = StackVec::new();
        for &b in &address.as_slice()[7..13] {
            if b != 32 {
                source.push(b)?;
            }
        }

        let mut info = StackVec::new();
        for &b in iter.as_ref() {
            info.push(b)?;
        }

        Some(Self { source, dest, info })
    }

    pub fn source(&self) -> &[u8] {
        self.source.as_slice()
    }

    pub fn dest(&self) -> &[u8] {
        self.dest.as_slice()
    }

    pub fn info(&self) -> &[u8] {
        self.info.as_slice()
    }
}
