use crate::queue::Queue;

use goertzel_nostd::Parameters;

pub(crate) struct BitDecoder<const N: usize> {
    sample_rate: u32,
    baud: u32,
    pll: f32,
    samples: Queue<N, f32>,
    prev_nrz: f32,

    upper: Parameters,
    lower: Parameters,
}

impl BitDecoder<{ 9600 / 1200 }> {
    pub(crate) fn new_9600_samples_1200_baud() -> Self {
        Self {
            sample_rate: 9600,
            baud: 1200,
            pll: 0.0,
            samples: Queue::new(),
            prev_nrz: 0.0,

            lower: Parameters::new(1200.0, 9600, 9600 / 1200),
            upper: Parameters::new(2200.0, 9600, 9600 / 1200),
        }
    }
}

impl<const N: usize> BitDecoder<N> {
    pub(crate) fn push(&mut self, sample: f32) -> Option<bool> {
        // slide our window
        self.samples.push(sample);

        self.sample().map(|x| x < 0.0)
    }

    fn sample(&mut self) -> Option<f32> {
        let value = self.decode_window();

        const A: f32 = 0.75;
        let zero_crossing =
            (self.prev_nrz > 0.0 && value < 0.0) || (self.prev_nrz < 0.0 && value > 0.0);
        if zero_crossing {
            self.pll *= A;
        }

        self.pll += self.baud as f32 / self.sample_rate as f32;
        let overflow = self.pll > 0.5;

        self.prev_nrz = value;
        if overflow {
            self.pll -= 1.0;
            Some(value)
        } else {
            None
        }
    }

    fn decode_window(&mut self) -> f32 {
        let a = self.samples.contiguous();

        // Goertzel breaks down with large amounts of DC offset. Let's remove it first
        let avg = a.iter().sum::<f32>() / N as f32;

        self.lower.mag_squared(a.iter().map(|x| x - avg))
            - self.upper.mag_squared(a.iter().map(|x| x - avg))
    }
}
