mod bit;
mod nrzi;
mod raw_frame;
mod unstuffer;

use crate::hdlc::Hdlc;
use crate::Ax25Frame;
use bit::BitDecoder;
use nrzi::Nrzi;
use raw_frame::RawFrame;

pub struct Decoder<const MAX_FRAME_SIZE: usize, const N: usize> {
    bit: BitDecoder<N>,
    nrzi: Nrzi,
    raw_frame: RawFrame<MAX_FRAME_SIZE>,
}

impl<const MAX_FRAME_SIZE: usize> Decoder<MAX_FRAME_SIZE, { 9600 / 1200 }> {
    pub fn new_9600_samples_1200_baud() -> Self {
        Self {
            bit: BitDecoder::new_9600_samples_1200_baud(),
            nrzi: Nrzi::new(),
            raw_frame: RawFrame::new(),
        }
    }

    pub fn receiving_frame(&self) -> bool {
        self.raw_frame.within_frame()
    }

    pub fn push(&mut self, sample: f32) -> Option<Ax25Frame<MAX_FRAME_SIZE>> {
        let bit = self.bit.push(sample)?;
        let bit = self.nrzi.push(bit);
        let (frame_buf, frame_len) = self.raw_frame.push(bit)?;
        let bytes = unstuffer::unstuff_bits(frame_buf, frame_len)?;
        let hdlc = Hdlc::decode(bytes)?;
        let ax25 = Ax25Frame::decode_raw(hdlc.unscrambled_data().as_slice());

        ax25
    }
}

#[cfg(test)]
mod tests {
    extern crate std;

    use super::*;
    use std::fs::File;

    #[test]
    fn decode_works() {
        let (header, samples) =
            wav::read(&mut File::open("test_data/packets.wav").unwrap()).unwrap();

        assert_eq!(9600, header.sampling_rate);

        let samples = match samples {
            wav::BitDepth::Sixteen(x) => x,
            _ => panic!("Wrong wav file type"),
        };

        let mut decoder: Decoder<1000, 8> = Decoder::new_9600_samples_1200_baud();

        let mut frames = std::vec::Vec::new();
        for s in samples {
            if let Some(f) = decoder.push(s as f32 / i16::MAX as f32) {
                frames.push(f);
            }
        }

        assert_eq!(
            &[
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  1 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  2 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  3 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  4 of 4"[..]
                )
                .unwrap()
            ],
            frames.as_slice()
        );
    }

    #[test]
    fn decode_works_with_dc_offset() {
        let (header, samples) =
            wav::read(&mut File::open("test_data/packets.wav").unwrap()).unwrap();

        assert_eq!(9600, header.sampling_rate);

        let samples = match samples {
            wav::BitDepth::Sixteen(x) => x,
            _ => panic!("Wrong wav file type"),
        };

        let mut decoder: Decoder<1000, 8> = Decoder::new_9600_samples_1200_baud();

        let mut frames = std::vec::Vec::new();
        for s in samples {
            let mut s = s as f32 / i16::MAX as f32;
            assert!(s < 0.5 && s > -0.5);
            s += 5.0;
            if let Some(f) = decoder.push(s) {
                frames.push(f);
            }
        }

        assert_eq!(
            &[
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  1 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  2 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  3 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  4 of 4"[..]
                )
                .unwrap()
            ],
            frames.as_slice()
        );
    }

    #[test]
    fn test_with_noise_before() {
        let (header, samples) =
            wav::read(&mut File::open("test_data/packets_with_noise_before.wav").unwrap()).unwrap();

        assert_eq!(9600, header.sampling_rate);

        let samples = match samples {
            wav::BitDepth::Sixteen(x) => x,
            _ => panic!("Wrong wav file type"),
        };

        let mut decoder: Decoder<1000, 8> = Decoder::new_9600_samples_1200_baud();

        let mut frames = std::vec::Vec::new();
        for s in samples {
            if let Some(f) = decoder.push(s as f32 / i16::MAX as f32) {
                frames.push(f);
            }
        }

        assert_eq!(
            &[
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  1 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  2 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  3 of 4"[..]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"WB2OSZ"[..],
                    &b"TEST"[..],
                    &b"\x03\xF0,The quick brown fox jumps over the lazy dog!  4 of 4"[..]
                )
                .unwrap()
            ],
            frames.as_slice()
        );
    }

    // todo
    #[test]
    fn actual_aprs_recording() {
        let (header, samples) =
            wav::read(&mut File::open("test_data/live_recording_from_aprs.wav").unwrap()).unwrap();

        assert_eq!(9600, header.sampling_rate);

        let samples = match samples {
            wav::BitDepth::Sixteen(x) => x,
            _ => panic!("Wrong wav file type"),
        };

        let mut decoder: Decoder<1000, 8> = Decoder::new_9600_samples_1200_baud();

        let mut frames = std::vec::Vec::new();
        for s in samples {
            let v = s as f32 / i16::MAX as f32;
            if let Some(f) = decoder.push(v) {
                frames.push(f);
            }
        }

        assert_eq!(
            &[
                Ax25Frame::new(
                    &b"VE9SDH"[..],
                    &b"T7PQ7P"[..],
                    &[
                        138, 98, 156, 174, 134, 226, 174, 146, 136, 138, 98, 64, 224, 172, 138,
                        114, 142, 140, 146, 228, 174, 146, 136, 138, 100, 64, 225, 3, 240, 96, 93,
                        56, 49, 109, 43, 68, 107, 47, 96, 34, 51, 103, 125, 86, 69, 57, 83, 68, 72,
                        32, 64, 32, 86, 69, 57, 83, 77, 84, 49, 52, 54, 46, 54, 49, 48, 95, 49, 13
                    ]
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"W1LH"[..],
                    &b"APU25N"[..],
                    &[
                        98, 152, 144, 64, 64, 242, 172, 138, 114, 140, 160, 142, 228, 174, 146,
                        136, 138, 102, 64, 99, 3, 240, 61, 52, 52, 52, 52, 46, 48, 53, 78, 47, 48,
                        54, 55, 51, 48, 46, 49, 52, 87, 114, 87, 65, 76, 90, 32, 66, 114, 111, 97,
                        100, 99, 97, 115, 116, 32, 115, 105, 116, 101, 32, 123, 85, 73, 86, 51, 50,
                        78, 125, 13
                    ],
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"W1LH"[..],
                    &b"APU25N"[..],
                    &[
                        98, 152, 144, 64, 64, 242, 172, 138, 114, 140, 160, 142, 228, 174, 146,
                        136, 138, 102, 64, 99, 3, 240, 62, 49, 51, 49, 56, 50, 50, 122, 85, 73, 45,
                        86, 105, 101, 119, 51, 50, 32, 86, 50, 46, 48, 51, 13
                    ],
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"W1LH"[..],
                    &b"APU25N"[..],
                    &[
                        98, 152, 144, 64, 64, 242, 172, 138, 114, 140, 160, 142, 228, 172, 138,
                        114, 178, 140, 134, 225, 3, 240, 61, 52, 52, 52, 52, 46, 48, 53, 78, 47,
                        48, 54, 55, 51, 48, 46, 49, 52, 87, 114, 87, 65, 76, 90, 32, 66, 114, 111,
                        97, 100, 99, 97, 115, 116, 32, 115, 105, 116, 101, 32, 123, 85, 73, 86, 51,
                        50, 78, 125, 13
                    ],
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"W1LH"[..],
                    &b"APU25N"[..],
                    &[
                        98, 152, 144, 64, 64, 242, 172, 138, 114, 140, 160, 142, 228, 172, 138,
                        114, 178, 140, 134, 225, 3, 240, 62, 49, 51, 49, 56, 50, 50, 122, 85, 73,
                        45, 86, 105, 101, 119, 51, 50, 32, 86, 50, 46, 48, 51, 13
                    ],
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"W1LH"[..],
                    &b"APU25N"[..],
                    &[
                        98, 152, 144, 64, 64, 242, 172, 138, 114, 140, 160, 142, 228, 172, 138,
                        114, 140, 160, 142, 224, 174, 146, 136, 138, 102, 64, 97, 3, 240, 61, 52,
                        52, 52, 52, 46, 48, 53, 78, 47, 48, 54, 55, 51, 48, 46, 49, 52, 87, 114,
                        87, 65, 76, 90, 32, 66, 114, 111, 97, 100, 99, 97, 115, 116, 32, 115, 105,
                        116, 101, 32, 123, 85, 73, 86, 51, 50, 78, 125, 13
                    ],
                )
                .unwrap(),
                Ax25Frame::new(
                    &b"W1LH"[..],
                    &b"APU25N"[..],
                    &[
                        98, 152, 144, 64, 64, 242, 172, 138, 114, 140, 160, 142, 228, 172, 138,
                        114, 140, 160, 142, 224, 174, 146, 136, 138, 102, 64, 97, 3, 240, 62, 49,
                        51, 49, 56, 50, 50, 122, 85, 73, 45, 86, 105, 101, 119, 51, 50, 32, 86, 50,
                        46, 48, 51, 13
                    ],
                )
                .unwrap(),
            ],
            frames.as_slice()
        );
    }
}
