use bitvec::prelude::*;

pub(crate) struct RawFrame<const MAX_FRAME_SIZE: usize> {
    buffer: u8,
    current_frame: BitArray<[u8; MAX_FRAME_SIZE], Msb0>,
    current_frame_len: usize,
    in_frame: bool,
}

impl<const MAX_FRAME_SIZE: usize> RawFrame<MAX_FRAME_SIZE> {
    pub(crate) fn new() -> Self {
        Self {
            buffer: 0,
            current_frame: BitArray::new([0; MAX_FRAME_SIZE]),
            current_frame_len: 0,
            in_frame: false,
        }
    }

    pub(crate) fn within_frame(&self) -> bool {
        // min frame length of 15. TODO shouldn't hardcode this
        self.in_frame && self.current_frame_len > 8 * 15
    }

    pub(crate) fn push(
        &mut self,
        bit: bool,
    ) -> Option<(BitArray<[u8; MAX_FRAME_SIZE], Msb0>, usize)> {
        let b = if bit { 1 } else { 0 };

        if self.buffer == 0x7e {
            // start of frame
            self.in_frame = true;
        }

        if self.in_frame {
            if self.current_frame_len >= MAX_FRAME_SIZE * 8 {
                // frame is too big. Give up
                self.in_frame = false;
                self.current_frame_len = 0;
            } else {
                self.current_frame.set(self.current_frame_len, bit);
                self.current_frame_len += 1;
            }
        }

        self.buffer <<= 1;
        self.buffer |= b;

        if self.buffer == 0x7e && self.in_frame {
            if self.current_frame_len >= 8 {
                // remove preamble
                self.current_frame_len -= 8;

                let ret = Some((self.current_frame, self.current_frame_len));

                // reset things
                self.current_frame_len = 0;

                return ret;
            } else {
                self.current_frame_len = 0;
            }
        }

        None
    }
}
