#[derive(Clone)]
pub struct Queue<const N: usize, T> {
    data: [T; N],
    end: usize,
}

impl<const N: usize, T: Copy + Default> Queue<N, T> {
    pub fn new() -> Self {
        Self {
            data: [T::default(); N],
            end: 0,
        }
    }

    pub fn push(&mut self, v: T) {
        self.data[self.end] = v;
        self.end = (self.end + 1) % N;
    }

    pub fn contiguous(&mut self) -> &mut [T; N] {
        // todo ideally we wouldn't need the default here? if possible
        let mut data = [T::default(); N];

        #[allow(clippy::needless_range_loop)]
        for i in 0..N {
            data[i] = self.data[(self.end + i) % N];
        }
        self.data = data;
        self.end = 0;

        &mut self.data
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn it_works() {
        let mut q = Queue::<4, u16>::new();
        assert_eq!(&[0, 0, 0, 0], q.contiguous());
        q.push(1);
        assert_eq!(&[0, 0, 0, 1], q.contiguous());
        q.push(2);
        assert_eq!(&[0, 0, 1, 2], q.contiguous());
        q.push(3);
        assert_eq!(&[0, 1, 2, 3], q.contiguous());
        q.push(4);
        assert_eq!(&[1, 2, 3, 4], q.contiguous());
        q.push(5);
        assert_eq!(&[2, 3, 4, 5], q.contiguous());
        q.push(6);
        assert_eq!(&[3, 4, 5, 6], q.contiguous());
        q.push(7);
        assert_eq!(&[4, 5, 6, 7], q.contiguous());
        q.push(8);
        assert_eq!(&[5, 6, 7, 8], q.contiguous());
        q.push(9);
        assert_eq!(&[6, 7, 8, 9], q.contiguous());
        q.push(10);
        assert_eq!(&[7, 8, 9, 10], q.contiguous());
    }

    #[test]
    fn it_works_without_making_contiguous_in_between() {
        let mut q = Queue::<4, u16>::new();
        assert_eq!(&[0, 0, 0, 0], q.clone().contiguous());
        q.push(1);
        assert_eq!(&[0, 0, 0, 1], q.clone().contiguous());
        q.push(2);
        assert_eq!(&[0, 0, 1, 2], q.clone().contiguous());
        q.push(3);
        assert_eq!(&[0, 1, 2, 3], q.clone().contiguous());
        q.push(4);
        assert_eq!(&[1, 2, 3, 4], q.clone().contiguous());
        q.push(5);
        assert_eq!(&[2, 3, 4, 5], q.clone().contiguous());
        q.push(6);
        assert_eq!(&[3, 4, 5, 6], q.clone().contiguous());
        q.push(7);
        assert_eq!(&[4, 5, 6, 7], q.clone().contiguous());
        q.push(8);
        assert_eq!(&[5, 6, 7, 8], q.clone().contiguous());
        q.push(9);
        assert_eq!(&[6, 7, 8, 9], q.clone().contiguous());
        q.push(10);
        assert_eq!(&[7, 8, 9, 10], q.clone().contiguous());
    }
}
