pub(crate) struct StackVec<const N: usize, T> {
    data: [T; N],
    count: usize,
}

impl<const N: usize, T: Default> StackVec<N, T> {
    pub fn new() -> Self {
        Self {
            data: core::array::from_fn(|_| T::default()),
            count: 0,
        }
    }

    pub fn len(&self) -> usize {
        self.count
    }

    #[must_use]
    pub fn push(&mut self, v: T) -> Option<()> {
        if self.count >= N {
            return None;
        }

        self.data[self.count] = v;
        self.count += 1;

        Some(())
    }

    pub fn pop(&mut self) -> Option<T> {
        if self.count == 0 {
            return None;
        }

        self.count -= 1;

        let mut val = T::default();
        core::mem::swap(&mut val, &mut self.data[self.count]);

        Some(val)
    }

    pub fn as_slice(&self) -> &[T] {
        &self.data[0..self.count]
    }

    pub fn as_slice_mut(&mut self) -> &mut [T] {
        &mut self.data[0..self.count]
    }
}

impl<const N: usize, T: core::fmt::Debug> core::fmt::Debug for StackVec<N, T> {
    fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
        write!(f, "{:?}", &self.data[0..self.count])
    }
}

impl<const N: usize> TryFrom<&[u8]> for StackVec<N, u8> {
    type Error = ();

    fn try_from(data: &[u8]) -> Result<Self, ()> {
        let mut out = Self::new();

        for &b in data {
            out.push(b).ok_or(())?;
        }

        Ok(out)
    }
}

impl<const N: usize, T: Default + PartialEq> PartialEq for StackVec<N, T> {
    fn eq(&self, other: &Self) -> bool {
        self.as_slice() == other.as_slice()
    }
}
impl<const N: usize, T: Default + Eq> Eq for StackVec<N, T> {}
